@@ -13,6 +13,8 @@
  */
 
 #include <map>
+
+#include <stdlib.h>
 #include <string.h>
 #include <jni.h>
 
@@ -60,13 +62,17 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
     if (!mutex)
     {
         mutex = new CRITICAL_SECTION();
-        InitializeCriticalSection(mutex);
+        InitializeCriticalSection(mutex);  // recursive by default
     }
 #else
     if (!mutex)
     {
+        pthread_mutexattr_t attr;
+
+        pthread_mutexattr_init(&attr);
+        pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
         mutex = new pthread_mutex_t();
-        pthread_mutex_init(mutex, NULL);
+        pthread_mutex_init(mutex, &attr);
     }
 #endif
 
@@ -89,6 +95,15 @@ void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
     _thr = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/RuntimeException"));
 #endif
 
+    _boo = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Boolean"));
+    _byt = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Byte"));
+    _cha = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Character"));
+    _dou = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Double"));
+    _flo = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Float"));
+    _int = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Integer"));
+    _lon = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Long"));
+    _sho = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Short"));
+
     _mids = new jmethodID[max_mid];
 
     _mids[mid_sys_identityHashCode] =
@@ -110,7 +125,6 @@ void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
         vm_env->GetMethodID(_obj, "getClass",
                             "()Ljava/lang/Class;");
 
-
     jclass iterable = vm_env->FindClass("java/lang/Iterable");
 
     if (iterable == NULL) /* JDK < 1.5 */
@@ -129,35 +143,43 @@ void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
                                 "next", "()Ljava/lang/Object;");
     }
 
-
     _mids[mid_enumeration_nextElement] =
         vm_env->GetMethodID(vm_env->FindClass("java/util/Enumeration"),
                             "nextElement", "()Ljava/lang/Object;");
 
     _mids[mid_Boolean_booleanValue] =
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Boolean"),
-                            "booleanValue", "()Z");
+        vm_env->GetMethodID(_boo, "booleanValue", "()Z");
     _mids[mid_Byte_byteValue] = 
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Byte"),
-                            "byteValue", "()B");
+        vm_env->GetMethodID(_byt, "byteValue", "()B");
     _mids[mid_Character_charValue] =
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Character"),
-                            "charValue", "()C");
+        vm_env->GetMethodID(_cha, "charValue", "()C");
     _mids[mid_Double_doubleValue] = 
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Double"),
-                            "doubleValue", "()D");
+        vm_env->GetMethodID(_dou, "doubleValue", "()D");
     _mids[mid_Float_floatValue] =
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Float"),
-                            "floatValue", "()F");
+        vm_env->GetMethodID(_flo, "floatValue", "()F");
     _mids[mid_Integer_intValue] = 
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Integer"),
-                            "intValue", "()I");
+        vm_env->GetMethodID(_int, "intValue", "()I");
     _mids[mid_Long_longValue] = 
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Long"),
-                            "longValue", "()J");
+        vm_env->GetMethodID(_lon, "longValue", "()J");
     _mids[mid_Short_shortValue] = 
-        vm_env->GetMethodID(vm_env->FindClass("java/lang/Short"),
-                            "shortValue", "()S");
+        vm_env->GetMethodID(_sho, "shortValue", "()S");
+
+    _mids[mid_Boolean_init] =
+        vm_env->GetMethodID(_boo, "<init>", "(Z)V");
+    _mids[mid_Byte_init] = 
+        vm_env->GetMethodID(_byt, "<init>", "(B)V");
+    _mids[mid_Character_init] =
+        vm_env->GetMethodID(_cha, "<init>", "(C)V");
+    _mids[mid_Double_init] = 
+        vm_env->GetMethodID(_dou, "<init>", "(D)V");
+    _mids[mid_Float_init] = 
+        vm_env->GetMethodID(_flo, "<init>", "(F)V");
+    _mids[mid_Integer_init] =
+        vm_env->GetMethodID(_int, "<init>", "(I)V");
+    _mids[mid_Long_init] = 
+        vm_env->GetMethodID(_lon, "<init>", "(J)V");
+    _mids[mid_Short_init] = 
+        vm_env->GetMethodID(_sho, "<init>", "(S)V");
 }
 
 int JCCEnv::attachCurrentThread(char *name, int asDaemon)
@@ -227,7 +249,7 @@ jobject JCCEnv::enumerationNext(jobject obj) const
 
 jboolean JCCEnv::isInstanceOf(jobject obj, getclassfn initializeClass) const
 {
-    return get_vm_env()->IsInstanceOf(obj, (*initializeClass)());
+    return get_vm_env()->IsInstanceOf(obj, getClass(initializeClass));
 }
 
 jclass JCCEnv::findClass(const char *className) const
@@ -239,7 +261,11 @@ jclass JCCEnv::findClass(const char *className) const
         JNIEnv *vm_env = get_vm_env();
 
         if (vm_env)
+        {
             cls = vm_env->FindClass(className);
+            if (cls == NULL)
+                reportException();
+        }
 #ifdef PYTHON
         else
         {
@@ -368,10 +394,23 @@ jobject JCCEnv::deleteGlobalRef(jobject obj, int id)
     return NULL;
 }
 
+jclass JCCEnv::getClass(getclassfn initializeClass) const
+{
+    jclass cls = (*initializeClass)(true);
+
+    if (cls == NULL)
+    {
+        lock locked;
+        cls = (*initializeClass)(false);
+    }
+
+    return cls;
+}
+
 jobject JCCEnv::newObject(getclassfn initializeClass, jmethodID **mids,
                           int m, ...)
 {
-    jclass cls = (*initializeClass)();
+    jclass cls = getClass(initializeClass);
     JNIEnv *vm_env = get_vm_env();
     jobject obj;
 
@@ -437,6 +476,48 @@ jclass JCCEnv::getPythonExceptionClass() const
 {
     return _thr;
 }
+
+// returns true if Python exception instance was successfully restored
+bool JCCEnv::restorePythonException(jthrowable throwable) const
+{
+#ifdef _jcc_lib   // PythonException is only available in shared mode
+    jclass pycls = getPythonExceptionClass();
+    JNIEnv *vm_env = get_vm_env();
+
+    // Support through-layer exceptions by taking the active PythonException
+    // and making the enclosed exception visible to Python again.
+
+    if (vm_env->IsSameObject(vm_env->GetObjectClass(throwable), pycls))
+    {
+        jfieldID fid = vm_env->GetFieldID(pycls, "py_error_state", "J");
+        PyObject *state = (PyObject *) vm_env->GetLongField(throwable, fid);
+        
+        if (state != NULL)
+        {
+            PyObject *type = PyTuple_GET_ITEM(state, 0);
+            PyObject *value = PyTuple_GET_ITEM(state, 1);
+            PyObject *tb = PyTuple_GET_ITEM(state, 2);
+
+            Py_INCREF(type);
+            if (value == Py_None)
+                value = NULL;
+            else
+                Py_INCREF(value);
+            if (tb == Py_None)
+                tb = NULL;
+            else
+                Py_INCREF(tb);
+
+            PyErr_Restore(type, value, tb);
+
+            return true;
+        }
+    }
+#endif
+
+    return false;
+}
+
 #endif
 
 void JCCEnv::reportException() const
@@ -634,6 +715,46 @@ jshort JCCEnv::shortValue(jobject obj) const
     return get_vm_env()->CallShortMethod(obj, _mids[mid_Short_shortValue]);
 }
 
+jobject JCCEnv::boxBoolean(jboolean value) const
+{
+    return get_vm_env()->NewObject(_boo, _mids[mid_Boolean_init], value);
+}
+
+jobject JCCEnv::boxByte(jbyte value) const
+{
+    return get_vm_env()->NewObject(_byt, _mids[mid_Byte_init], value);
+}
+
+jobject JCCEnv::boxChar(jchar value) const
+{
+    return get_vm_env()->NewObject(_cha, _mids[mid_Character_init], value);
+}
+
+jobject JCCEnv::boxDouble(jdouble value) const
+{
+    return get_vm_env()->NewObject(_dou, _mids[mid_Double_init], value);
+}
+
+jobject JCCEnv::boxFloat(jfloat value) const
+{
+    return get_vm_env()->NewObject(_flo, _mids[mid_Float_init], value);
+}
+
+jobject JCCEnv::boxInteger(jint value) const
+{
+    return get_vm_env()->NewObject(_int, _mids[mid_Integer_init], value);
+}
+
+jobject JCCEnv::boxLong(jlong value) const
+{
+    return get_vm_env()->NewObject(_lon, _mids[mid_Long_init], value);
+}
+
+jobject JCCEnv::boxShort(jshort value) const
+{
+    return get_vm_env()->NewObject(_sho, _mids[mid_Short_init], value);
+}
+
 
 jmethodID JCCEnv::getMethodID(jclass cls, const char *name,
                               const char *signature) const
@@ -744,10 +865,10 @@ void JCCEnv::setClassPath(const char *classPath)
     jmethodID mu = vm_env->GetMethodID(_fil, "toURL", "()Ljava/net/URL;");
     jmethodID ma = vm_env->GetMethodID(_ucl, "addURL", "(Ljava/net/URL;)V");
 #if defined(_MSC_VER) || defined(__WIN32)
-    char *pathsep = ";";
+    const char *pathsep = ";";
     char *path = _strdup(classPath);
 #else
-    char *pathsep = ":";
+    const char *pathsep = ":";
     char *path = strdup(classPath);
 #endif
 
@@ -774,9 +895,9 @@ char *JCCEnv::getClassPath()
     jmethodID gu = vm_env->GetMethodID(_ucl, "getURLs", "()[Ljava/net/URL;");
     jmethodID gp = vm_env->GetMethodID(_url, "getPath", "()Ljava/lang/String;");
 #if defined(_MSC_VER) || defined(__WIN32)
-    char *pathsep = ";";
+    const char *pathsep = ";";
 #else
-    char *pathsep = ":";
+    const char *pathsep = ":";
 #endif
     jobjectArray array = (jobjectArray)
         vm_env->CallObjectMethod(classLoader, gu);
@@ -839,9 +960,26 @@ char *JCCEnv::toUTF(jstring str) const
 
 char *JCCEnv::toString(jobject obj) const
 {
-    return obj
-        ? toUTF((jstring) callObjectMethod(obj, _mids[mid_obj_toString]))
-        : NULL;
+    try {
+        return obj
+            ? toUTF((jstring) callObjectMethod(obj, _mids[mid_obj_toString]))
+            : NULL;
+    } catch (int e) {
+        switch (e) {
+          case _EXC_PYTHON:
+            return NULL;
+          case _EXC_JAVA: {
+              JNIEnv *vm_env = get_vm_env();
+
+              vm_env->ExceptionDescribe();
+              vm_env->ExceptionClear();
+
+              return NULL;
+          }
+          default:
+            throw;
+        }
+    }
 }
 
 char *JCCEnv::getClassName(jobject obj) const
