@@ -1355,10 +1355,13 @@ PyObject *PyErr_SetJavaError()
 {
     JNIEnv *vm_env = env->get_vm_env();
     jthrowable throwable = vm_env->ExceptionOccurred();
-    PyObject *err;
 
     vm_env->ExceptionClear();
-    err = t_Throwable::wrap_Object(Throwable(throwable));
+
+    if (env->restorePythonException(throwable))
+        return NULL;
+
+    PyObject *err = t_Throwable::wrap_Object(Throwable(throwable));
 
     PyErr_SetObject(PyExc_JavaError, err);
     Py_DECREF(err);
